// License: GPL version 2 or higher. See http://www.gnu.org/licenses/gpl.html
// Written by Freek van Walderveen <freek -at- vanwal -dot- nl>, 2008.

#include <proj_api.h>
#include <limits.h>

#include <QXmlStreamReader>
#include <QFile>
#include <QImage>
#include <QColor>
#include <QDateTime>
#include <iostream>

//TODO make these defines into runtime options
// Define this to calculate bounding box of input instead of making an image.
#undef CALC_BBOX

#undef COLOUR_AGE

double stringRefToDouble(QStringRef s, bool *ok) {
  if (s.string()) {
    return s.toString().toDouble(ok);
  } else {
    if (ok) {
      *ok = false;
    }
    return 0;
  }
}

struct areaInfo {
#ifdef COLOUR_AGE
  uint timestamp;
#else
  int nodeCountUser;
  int nodeCountOthers;
#endif
};

int main(int argc, char **argv) {
  // Pick your favourite projection here:
  // From http://lists.maptools.org/pipermail/proj/2006-March/002168.html
  /*
  const char *proj_arg =  "+proj=latlong +ellps=WGS84 +to +proj=sterea +lat_0=52.15616055555555"
                    "+lon_0=5.38763888888889 +k=0.9999079 +x_0=155000 +y_0=463000"
                    "+ellps=bessel +units=m +no_defs"
  */
  /*
  const char *proj_arg =
                  "+proj=latlong +datum=WGS84 +to +proj=sterea +lat_0=52.15616055555555 "
                  "+lon_0=5.38763888888889 +k=0.999908 +x_0=155000 +y_0=463000 "
                  "+ellps=bessel +units=m +no_defs "
                  "+towgs84=565.2369,50.0087,465.658,-0.406857330322398,0.350732676542563,-1.8703473836068,4.0812";
  */
  const char *proj_arg = 
      "+proj=merc +a=6378137 +b=6378137 +lat_ts=0.0 +lon_0=0.0 +x_0=0.0 +y_0=0 +k=1.0 +units=m +nadgrids=@null +no_defs +over";
  projUV p;
  projPJ pj;

  if (!(pj = pj_init_plus(proj_arg))) {
    std::cerr << "Failed initializing proj4." << std::endl;
    return 1;
  }
  
//  double sqrScale = scale * scale * 10000; // = scale^2 / (1/100)^2
#ifdef CALC_BBOX
  if (argc != 2) {
    std::cerr << "Usage: " << argv[0] << " <scale>" << std::endl;
    std::cerr << "<scale>^2 square metre will be covered by one pixel." << std::endl;
    return 1;
  }
  double scale = 1.0/strtod(argv[1],0);
  int uMin = INT_MAX, uMax = INT_MIN, vMin = INT_MAX, vMax = INT_MIN;
#else
  if (argc != 6) {
    std::cerr << "Usage: " << argv[0] << " <scale> <min x> <max x> <min y> <max y>" << std::endl;
    std::cerr << "<scale>^2 square metre will be covered by one pixel." << std::endl;
    std::cerr << "The bounding box coordinates are measured in pixels, i.e. <metres from the origin> * <scale>." << std::endl;
    return 1;
  }
  double scale = 1.0/strtod(argv[1],0);
  // NL bbox for scale=1/100: [3115:8224] x [65554:71275]
  const int uMin = atoi(argv[2]),
            uMax = atoi(argv[3]),
            vMin = atoi(argv[4]),
            vMax = atoi(argv[5]);
  
  const int width =  uMax - uMin + 1,
      height = vMax - vMin + 1;
  
  areaInfo **areas = new areaInfo*[width];
  for (int u = 0; u < width; ++u) {
    areas[u] = new areaInfo[height];
    for (int v = 0; v < height; ++v) {
#ifdef COLOUR_AGE
      areas[u][v].timestamp = 0;
#else
      areas[u][v].nodeCountUser = 0;
      areas[u][v].nodeCountOthers = 0;
#endif
    }
  }
#endif
  
  // Count number of nodes for each bucket
  QFile file;
  file.open(stdin, QIODevice::ReadOnly);
  QXmlStreamReader xml(&file);
  while (!xml.atEnd()) {
    if (xml.readNext() == QXmlStreamReader::StartElement) {
      if (xml.name() == "node") {
        bool ok;
        p.u = stringRefToDouble(xml.attributes().value("lon"), &ok);
        if (!ok) { continue; }
        p.v = stringRefToDouble(xml.attributes().value("lat"), &ok);
        if (!ok) { continue; }
        
        // Do the projection
        p.u *= DEG_TO_RAD;
        p.v *= DEG_TO_RAD;
        p = pj_fwd(p, pj);
        
        int u = static_cast<int>(p.u * scale);
        int v = static_cast<int>(p.v * scale);
        
#ifdef CALC_BBOX
        if (u < uMin) { uMin = u; }
        if (u > uMax) { uMax = u; }
        if (v < vMin) { vMin = v; }
        if (v > vMax) { vMax = v; }
#else
        if (u < uMin || u > uMax ||
            v < vMin || v > vMax) {
          continue;
        }
        
        u -= uMin;
        v -= vMin;
        
#ifdef COLOUR_AGE
        uint timestamp = QDateTime::fromString(
                xml.attributes().value("timestamp").toString(),
                Qt::ISODate).toTime_t();
        if (areas[u][v].timestamp < timestamp) {
          areas[u][v].timestamp = timestamp;
        }
#else
        if (xml.attributes().value("user") == "AND") {
          areas[u][v].nodeCountUser++;
        } else {
          areas[u][v].nodeCountOthers++;
        }
#endif

#endif
      }
    }
  }
  if (xml.hasError()) {
    std::cerr << qPrintable(xml.errorString()) << std::endl;
  }
  file.close();
  
#ifdef CALC_BBOX
  std::cout << "bbox: [" << uMin << ":" << uMax << "] x ["
                         << vMin << ":" << vMax << "]" << std::endl;
#else
  // Make the image
  int max = 0;
  QImage img(width, height, QImage::Format_RGB32);
  for (int u = 0; u < width; ++u) {
    for (int v = 0; v < height; ++v) {
      QColor c;
#ifdef COLOUR_AGE
      const uint finishAND = 1190643700; // approx. moment that AND import finished
      uint now = QDateTime::currentDateTime().toTime_t();
      uint timeSpan = (now - finishAND) * 3/2;
      float age = 1.0 * (now - areas[u][v].timestamp) / timeSpan;
      int r = qMin(255, qMax(0, static_cast<int>( age          * 3 * 255)));
      int g = qMin(255, qMax(0, static_cast<int>((age - 1.0/3) * 3 * 255)));
      int b = qMin(255, qMax(0, static_cast<int>((age - 2.0/3) * 3 * 255)));
      img.setPixel(u, height - v - 1, qRgb(r, g, b));
#else
      int total = areas[u][v].nodeCountUser + areas[u][v].nodeCountOthers;
      if (total > max) { max = total; }
      if (total == 0) {
        img.setPixel(u, height - v - 1, qRgb(0, 0, 0));
      } else {
        // Magic function ;-)
        float hue;
        if (areas[u][v].nodeCountOthers < 1) {
          hue = 1;
        } else if (areas[u][v].nodeCountUser < 1) {
          hue = 0;
        } else {
          hue = 0.5 + log(1.0 * total / areas[u][v].nodeCountOthers - 1.0)/10.0;
          if (hue > 1) { hue = 1; }
          if (hue < 0) { hue = 0; }
        }
        
        // Note: valueOffset + valueScale * log_e(scale*100*<max nodes/bucket>)
        //       should roughly be equal to 255.
        const int valueOffset = 100,
                  valueScale = 55;
        int value = qMax(0, qMin(255, 
                      static_cast<int>(log(scale * 100 * total) * valueScale + valueOffset)));
        c = QColor::fromHsv(static_cast<int>(240 * hue), 255, value);
        img.setPixel(u, height - v - 1, c.rgb());
      }
#endif
    }
  }
  QFile outFile;
  outFile.open(stdout, QIODevice::WriteOnly);
  img.save(&outFile, "PNG");
  outFile.close();
  
  std::cerr << "Max nodes/bucket: " << max << std::endl;
#endif
  
  return 0;
}
